/*******************************************************************************
 * Copyright (c) 2009 Information-technology Promotion Agency, Japan.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package benten.cat.tm.engine.omegat;

import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.omegat.core.data.LegacyTM;
import org.omegat.core.data.StringEntry;
import org.omegat.core.data.StringEntryWrapper;
import org.omegat.core.matching.FuzzyMatcherWrapper;
import org.omegat.core.matching.NearString;
import org.omegat.util.Language;
import org.omegat.util.TMXReader;

import benten.cat.tm.core.BentenTmEngine;
import benten.cat.tm.core.BentenTmSearchResult;
import benten.cat.tm.engine.omegat.messages.DefaultTmEngineMessages;
import blanco.xliff.BlancoXliffParser;
import blanco.xliff.valueobject.BlancoXliff;
import blanco.xliff.valueobject.BlancoXliffBody;
import blanco.xliff.valueobject.BlancoXliffFile;
import blanco.xliff.valueobject.BlancoXliffTransUnit;

/**
 * デフォルトの Benten TM (翻訳メモリ) エンジン。
 * 
 * ★基本設計「共通機能: 翻訳メモリー・エンジン機能」に対応します。
 */
public class DefaultTmEngine implements BentenTmEngine {
	/**
	 * Benten デフォルト TM エンジン・プラグインのためのメッセージ。
	 */
	protected static final DefaultTmEngineMessages fMsg = new DefaultTmEngineMessages();

	private static final boolean IS_DEBUG = Boolean.valueOf(System.getProperty("benten.debug")); //$NON-NLS-1$

	private static Pattern extensionPattern = Pattern.compile("\\.[^.]*$"); //$NON-NLS-1$

	private List<LegacyTM> legacyTmList = new ArrayList<LegacyTM>();

	private final List<BentenTmSearchResult> nearList = new ArrayList<BentenTmSearchResult>();

	/**
	 * コンストラクター。
	 */
	public DefaultTmEngine() {
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see benten.cat.tm.core.BentenTmEngine#loadTmx(java.io.File)
	 */
	public void loadTmx(File dirTmx) throws IOException {
		final List<LegacyTM> result = new ArrayList<LegacyTM>();

		final File[] fileTmxs = dirTmx.listFiles();
		if (fileTmxs == null) {
			throw new IllegalArgumentException(fMsg.getMsg0(dirTmx.toString()));
		}

		for (int index = 0; index < fileTmxs.length; index++) {
			final File fileTmx = fileTmxs[index];
			if (fileTmx.isFile() == false) {
				continue;
			}
			if (fileTmx.getName().toLowerCase().endsWith(".tmx") == false) { //$NON-NLS-1$
				// ファイルの拡張子が .tmx のものだけ処理します。
				continue;
			}
			// TMXReaderは、BOM付きのUTF-8またはUTF-16については、文字エンコーディングを自動判別します。
			// TMXReaderの第１引数に指定した文字エンコーディングは、自動判別できなかった場合の
			// デフォルトとして使用されます。
			final TMXReader reader = new TMXReader("UTF-8", new Language("en-US"), new Language("ja-JP"), false); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			reader.loadFile(fileTmx.getAbsolutePath(), true);
			final LegacyTM legacyTm = loadLegacyTM(stripExtension(fileTmx.getName()), reader);
			result.add(legacyTm);
		}
		if (result.size() > 0)
			legacyTmList = result;
	}

	/**
	 * TMXReader を使って、TMX ファイルを LegacyTM に読み込みます。
	 * 
	 * @param tmxFileName
	 *            reader にロードした TMX のファイル名。
	 * @param reader
	 *            TMX ファイルがロードされた TMXReader。
	 * @return TMX ファイルが読み込まれた LegacyTM
	 */
	private LegacyTM loadLegacyTM(final String tmxFileName, final TMXReader reader) {
		final int num = reader.numSegments();
		final List<StringEntry> strEntryList = new ArrayList<StringEntry>(num);
		for (int i = 0; i < num; i++) {
			final String src = reader.getSourceSegment(i);
			final String trans = reader.getTargetSegment(i);

			final StringEntryWrapper se = new StringEntryWrapper(src);
			se.setTranslation(trans);
			strEntryList.add(se);
		}

		return new LegacyTM(tmxFileName, strEntryList);
	}

	// TODO XLIFFをTMとして読み込む機能が不要なら、このメソッドは不要。
	public void loadLocalTmFromXliff(File fileXliff) {
		final List<LegacyTM> result = new ArrayList<LegacyTM>();

		final BlancoXliff xliff = new BlancoXliffParser().parse(fileXliff);
		final List<?> fileList = xliff.getFileList();

		for (int i = 0; i < fileList.size(); i++) {
			final BlancoXliffFile file = (BlancoXliffFile) fileList.get(i);
			final BlancoXliffBody body = file.getBody();
			final List<?> tuList = body.getTransUnitList();

			final List<StringEntry> tuEntry = new ArrayList<StringEntry>(tuList.size());
			for (int j = 0; j < tuList.size(); j++) {
				final BlancoXliffTransUnit tu = (BlancoXliffTransUnit) tuList.get(j);
				if (tu.getSource() == null) {
					continue;
				}
				if (tu.getTarget() == null) {
					continue;
				}
				if (tu.getSource().equals(tu.getTarget().getTarget())) {
					continue;
				}

				final StringEntryWrapper se = new StringEntryWrapper(tu.getSource());
				se.setTranslation(tu.getTarget().getTarget());
				tuEntry.add(se);
			}
			final List<StringEntry> localTmEntry = new ArrayList<StringEntry>();
			localTmEntry.addAll(tuEntry);
			final LegacyTM legacyTm = new LegacyTM(fileXliff.getName(), localTmEntry);
			result.add(legacyTm);
		}

		if (result.size() > 0)
			legacyTmList = result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see benten.cat.tm.core.BentenTmEngine#fuzzySearch(java.lang.String)
	 */
	public List<BentenTmSearchResult> fuzzySearch(String target) {
		nearList.clear();
		final FuzzyMatcherWrapper fuzzyMatcher = FuzzyMatcherWrapper.getInstance();

		final StringBuffer targetStrBuf = new StringBuffer();
		if (target != null) {
			targetStrBuf.append(target);
		}

		final StringEntry tuEntry = new StringEntry(targetStrBuf.toString());

		final List<NearString> nearStringList = fuzzyMatcher.match(tuEntry, legacyTmList);

		if (nearStringList.size() > 0) {
			if (IS_DEBUG) {
				System.out.println(fMsg.getMsg1(BigDecimal.valueOf(nearStringList.size())));
				System.out.println(fMsg.getMsg2() + tuEntry.getSrcText());
			}

			for (int i = 0; i < nearStringList.size(); i++) {
				final NearString nearString = nearStringList.get(i);
				final BentenTmSearchResult result = new BentenTmSearchResult();
				result.setMatchQuality(Integer.toString(nearString.score) + "% "); //$NON-NLS-1$
				result.setSource(nearString.str.getSrcText());
				result.setTarget(nearString.str.getTranslation());
				result.setOrigin(nearString.proj);
				nearList.add(result);

				if (IS_DEBUG) {
					System.out.println("------------------------------------"); //$NON-NLS-1$
					System.out.println(fMsg.getMsg3(BigDecimal.valueOf(i + 1)));
					System.out.println(fMsg.getMsg4(BigDecimal.valueOf(nearString.score)));
					System.out.println(nearString.str.getSrcText());
					System.out.println(nearString.str.getTranslation());
					System.out.println(nearString.proj);
				}
			}

			if (IS_DEBUG) {
				System.out.println("------------------------------------"); //$NON-NLS-1$
				System.out.println(""); //$NON-NLS-1$
			}
		}

		return nearList;

	}

	public void fuzzyMatchLocalTm(String target) {
		nearList.clear();
		final FuzzyMatcherWrapper fuzzyMatcher = FuzzyMatcherWrapper.getInstance();

		final StringBuffer targetStrBuf = new StringBuffer();
		if (target != null) {
			targetStrBuf.append(target);
		}
		final StringEntry tuEntry = new StringEntry(targetStrBuf.toString());

		final List<NearString> nearStringList = fuzzyMatcher.match(tuEntry, legacyTmList);

		if (nearStringList.size() > 0) {
			if (IS_DEBUG) {
				System.out.println(fMsg.getMsg5(BigDecimal.valueOf(nearStringList.size())));
				System.out.println(fMsg.getMsg6() + tuEntry.getSrcText());
			}

			for (int i = 0; i < nearStringList.size(); i++) {
				final NearString nearString = nearStringList.get(i);
				final BentenTmSearchResult result = new BentenTmSearchResult();
				result.setMatchQuality(Integer.toString(nearString.score) + "% "); //$NON-NLS-1$
				result.setSource(nearString.str.getSrcText());
				result.setTarget(nearString.str.getTranslation());
				result.setOrigin(nearString.proj);
				nearList.add(result);

				if (IS_DEBUG) {
					System.out.println("------------------------------------"); //$NON-NLS-1$
					System.out.println(fMsg.getMsg7(BigDecimal.valueOf(i + 1)));
					System.out.println(fMsg.getMsg8(BigDecimal.valueOf(nearString.score)));
					System.out.println(nearString.str.getSrcText());
					System.out.println(nearString.str.getTranslation());
				}
			}

			if (IS_DEBUG) {
				System.out.println("------------------------------------"); //$NON-NLS-1$
				System.out.println(""); //$NON-NLS-1$
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see benten.cat.tm.core.BentenTmEngine#unload()
	 */
	public void unload() {
		legacyTmList = new ArrayList<LegacyTM>();
	}

	/**
	 * ファイル名から拡張子を取り除きます。
	 * 
	 * @param target
	 *            ファイル名。
	 * @return 拡張子を取り除いたファイル名。
	 */
	private String stripExtension(String target) {
		if (target == null) {
			return target;
		}

		if (!target.contains(".")) { //$NON-NLS-1$
			return target;
		}

		final Matcher m = extensionPattern.matcher(target);
		m.find();
		return target.substring(0, m.start());
	}
}
